/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "all.h"
#include "cutils.h"
#include "keyword.h"

/*------------------------------------------------------------------*/
/*      Find the KEY WORD *cp, and return it's number               */
/*------------------------------------------------------------------*/
/*                                                                  */
/* You can add a function, you MUST place it in alphabetical order  */
/* and give it the next unused index number                         */
/*            *** 50 is RESERVED for assignment ***                 */
/* when adding you need to edit pass.cpp and run.cpp                */
/* CURRENT MAX IS 86 next command number is 87                      */
/*------------------------------------------------------------------*/

#define NKEYS (sizeof mkeywfn / sizeof(struct mkeyw))
struct mkeyw { const char *word; int index;  } mkeywfn[] = {
	{  "!",               53                        },
	{ "ABOUND",           GLE_KW_ABOUND             },
	{ "ALINE",            1                         },
	{ "AMOVE",            GLE_KW_AMOVE              },
	{ "ARC",              3                         },
	{ "ARCTO",            4                         },
	{ "ASETPOS",          73                        },
	{ "Assignment",       51                        },
	{ "BEGIN",            5                         },
	{ "BEZIER",           6                         },
	{ "BIGFILE",          58                        },
	{ "BITMAP",           68                        },
	{ "BITMAP_INFO",      69                        },
	{ "BOX",              7                         },
	{ "CALL",             52                        },
	{ "CIRCLE",           8                         },
	{ "CLOSEPATH",        9                         },
	{ "COLORMAP",         GLE_KW_COLORMAP           },
	{ "COMPATIBILITY",    GLE_KW_COMPATIBILITY      },
	{ "CURVE",            10                        },
	{ "DECLARE",          GLE_KW_DECLARESUB         },
	{ "DEFAULT",          GLE_KW_DEFAULT            },
	{ "DEFCOLOR",         GLE_KW_DEFCOLOR           },
	{ "DEFINE",           11                        },
	{ "DEFMARKER",        60                        },
	{ "DFONT",            12                        },
	{ "DRAW",             56                        },
	{ "ELLIPSE",          70                        },
	{ "ELLIPTICAL_ARC",   71                        },
	{ "ELLIPTICAL_NARC",  72                        },
	{ "ELSE",             13                        },
	{ "END",              14                        },
	{ "FCLOSE",           15                        },
	{ "FGETLINE",         75                        },
	{ "FILL",             16                        },
	{ "FOPEN",            17                        },
	{ "FOR",              18                        },
	{ "FREAD",            61                        },
	{ "FREADLN",          62                        },
	{ "FTOKENIZER",       76                        },
	{ "FWRITE",           63                        },
	{ "FWRITELN",         64                        },
	{ "GOTO",             19                        },
	{ "GRESTORE",         54                        },
	{ "GSAVE",            20                        },
	{ "ICON",             21                        },
	{ "IF",	              22                        },
	{ "INCLUDE",          23                        },
	{ "INPUT",            24                        },
	{ "JOIN",             25                        },
	{ "MARGINS",          78                        },
	{ "MARKER",           26                        },
	{ "MOVE",             27                        },
	{ "NAME",             39                        },
	{ "NARC",             28                        },
	{ "NEWPATH",          29                        },
	{ "NEXT",             30                        },
	{ "ORIENTATION",      79                        },
	{ "PAPERSIZE",        77                        },
	{ "PIE",              31                        },
	{ "PLOTTER",          57                        },
	{ "POSTSCRIPT",       55                        },
	{ "PRINT",            32                        },
	{ "PSBBTWEAK",        66                        },
	{ "PSCOMMENT",        65                        },
	{ "RBEZIER",          33                        },
	{ "REGION",           34                        },
	{ "RESTOREDEFAULTS",  GLE_KW_RESTOREDEFAULTS    },
	{ "RETURN",           50                        },
	{ "REVERSE",          35                        },
	{ "RLINE",            36                        },
	{ "RMOVE",            37                        },
	{ "ROT",              38                        },
	{ "ROTATE",           38                        },
	{ "RSETPOS",          81                        },
	{ "SAVE",             39                        },
	{ "SCALE",            40                        },
	{ "SET",              41                        },
	{ "SIZE",             42                        },
	{ "SLEEP",            GLE_KW_SLEEP,             },
	{ "STROKE",           43                        },
	{ "SUB",              44                        },
	{ "TEX",              74                        },
	{ "TEXT",             45                        },
	{ "TEXTDEF",          59                        },
	{ "TIFF",             67                        },
	{ "TRAN",             46                        },
	{ "TRANSLATE",        46                        },
	{ "UNTIL",            47                        },
	{ "WHILE",            48                        },
	{ "WRITE",            49                        }
};

int binsearchk(const char *word, struct mkeyw tab[], int n);

void cmd_name(int idx, char **cp) {
	static char *kp;
	static char fail[]="Keyword not found";
	if (kp==NULL) kp = (char*) myallocz(80);
	for (unsigned int i=0;i<NKEYS;i++) {
		if (mkeywfn[i].index==idx) {
			strcpy(kp,mkeywfn[i].word);
			*cp = kp;
			return;
		}
	}
	*cp = &fail[0];
}

void find_mkey(string cp, int *idx) {
	if (cp.length() == 0 ) { *idx = 0; return;}
	int i;
	i = binsearchk(cp.c_str(),mkeywfn,NKEYS);
	/* printf("I %d, cmd {%s} \n",i,cp); */
	if (i==-1) { *idx = 0; return;}
	*idx = mkeywfn[i].index;
}

/*------------------------------------------------------------------*/
/* Simple binary search 					    */
/*------------------------------------------------------------------*/

int binsearchk(const char *word, struct mkeyw tab[], int n) {
	int cond,low,high,mid;
	low = 0;
	high = n-1;
	while (low <= high) {
		mid = (low+high) / 2;
		if ((cond = strcmp(word,tab[mid].word)) < 0)
			high = mid - 1;
		else if (cond > 0)
			low = mid + 1;
		else
			return mid;
	}
	return -1;
}


/*-------------------------------------------------------------------------*/
/* This is for the tex primitives */

#include "tex.h"
struct mkeyw tkeywfn[] = {
	{ "^",                  tp_sup          },
	{ "_",                  tp_sub          },
	{ "acccmb",             tp_acccmb       },
	{ "accent",             tp_accent       },
	{ "accentxy",           tp_accentxy     },
	{ "char",               tp_char         },
	{ "chardef",            tp_chardef      },
	{ "def",                tp_def          },
	{ "defbegin",           tp_defbegin     },
	{ "delcode",            tp_delcode      },
	{ "delimiter",          tp_delimiter    },
	{ "frac",               tp_frac         },
	{ "hfill",              tp_hfill        },
	{ "left",               tp_left         },
	{ "linegap",            tp_linegap      },
	{ "lineskip",           tp_lineskip     },
	{ "mathchar",           tp_mathchar     },
	{ "mathchardef",        tp_mathchardef  },
	{ "mathcode",           tp_mathcode     },
	{ "movexy",             tp_movexy       },
	{ "newline",            tp_newline      },
	{ "nolimits",           tp_nolimits     },
	{ "overbrace",          tp_overbrace    },
	{ "overline",           tp_overline     },
	{ "parskip",            tp_parskip      },
	{ "presave",            tp_presave      },
	{ "right",              tp_right        },
	{ "rule",               tp_rule         },
	{ "setfont",            tp_setfont      },
	{ "sethei",             tp_sethei       },
	{ "setstretch",         tp_setstretch   },
	{ "sfont",              tp_sfont        },
	{ "ssfont",             tp_ssfont       },
	{ "sub",                tp_sub          },
	{ "sup",                tp_sup          },
	{ "tex",                tp_tex          },
	{ "tfont",              tp_tfont        },
	{ "underbrace",         tp_underbrace   },
	{ "underline",          tp_underline    },
	{ "unichr",             tp_unichr       },
	{ "unicode",            tp_unicode      }
};

#define NTKEYS (sizeof tkeywfn / sizeof(struct mkeyw))
int find_primcmd(char *cp) {
	int i;
	i = binsearchk(cp,tkeywfn,NTKEYS);
	if (i==-1) return 0;
	return tkeywfn[i].index;
}

